# OB - write your own commands.
#
# handler (hdl)

"event handler."

# imports

import importlib, importlib.util, os, queue, threading, zipfile, _thread

from olib import Default, Object, locked, update
from ob.tsk import launch
from ob.utl import direct, find_cmds, find_modules, get_exception

# defines

def __dir__():
    return ("Event", "Handler")

dispatchlock = _thread.allocate_lock()

# classes

class Event(Default):

    "basic event."

    def __init__(self):
        super().__init__()
        self.args = []
        self.cmd = ""
        self.ready = threading.Event()
        self.rest = ""
        self.result = []
        self.thrs = []
        self.txt = ""

    def parse(self):
        "parse command and arguments depending on self.txt."
        args = self.txt.split()
        if args:
            self.cmd = args[0]
        if len(args) >= 2:
            self.args = args[1:]
            self.rest = " ".join(args[1:])

    def reply(self, txt):
        "add txt to the result."
        if not self.result:
            self.result = []
        self.result.append(txt)

    def show(self):
        "show result."
        for txt in self.result:
            print(txt)

    def wait(self):
        "wait for event to finish."
        self.ready.wait()
        res = []
        for thr in self.thrs:
            res.append(thr.join())
        return res

class Handler(Object):

    "The Event Handler itself."

    def __init__(self):
        super().__init__()
        self.cmds = Object()
        self.queue = queue.Queue()
        self.speed = "fast"
        self.stopped = False

    def cmd(self, txt):
        "execute text on the handler."
        e = Event()
        e.txt = txt
        self.dispatch(e)
        return e

    @locked(dispatchlock)
    def dispatch(self, e):
        "dispatch an event to its registered command."
        e.parse()
        if e.cmd in self.cmds:
            try:
                self.cmds[e.cmd](e)
            except Exception as ex:
                print(get_exception(ex))
        e.show()
        e.ready.set()

    def handler(self):
        "loop for events."
        while not self.stopped:
            event = self.queue.get()
            if not event:
                break
            if not event.orig:
                event.orig = repr(self)
            event.speed = self.speed
            launch(self.dispatch, event)

    def load_mod(self, name):
        "load a module onto the handler."
        mod = direct(name)
        self.scan(mod)
        return mod

    def scan(self, mod):
        "scan a module for commands."
        cmds = find_cmds(mod)
        update(self.cmds, cmds)

    def start(self):
        "start the handler."
        launch(self.handler)

    def stop(self):
        "stop handler."
        self.stopped = True
        self.queue.put(None)

    def walk(self, names, skip=None):
        "walk a packages modules."
        modules = []
        if not skip:
            skipped = []
        else:
            skipped = skip.split(",")
        for name in names.split(","):
            spec = importlib.util.find_spec(name)
            if not spec:
                continue
            pkg = importlib.util.module_from_spec(spec)
            pn = getattr(pkg, "__path__", None)
            if not pn:
                continue
            pn = pn[0]
            if os.path.exists(pn):
                fls = os.listdir(pn)
            else:
                fls = zipfile.namelist(pn)
            for fn in fls:
                if fn.startswith("_") or not fn.endswith(".py"):
                    continue
                fn = fn[:-3]
                if fn in skipped:
                    continue
                mn = "%s.%s" % (name, fn)
                module = self.load_mod(mn)
                modules.append(module)
        return modules

# utilities

def direct(name):
    "do a direct import."
    return importlib.import_module(name)

# commands

def mods(event):
    "show loadable modules."
    event.reply(",".join([m.__name__.split(".")[-1] for m in find_modules("ob", "obmods")]))
